# coding=utf-8
# Copyright (c) 2015-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for compression in file backends."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import itertools
import os
import subprocess
try:
    import mock
except ImportError:
    from unittest import mock

import pytest
import six

from framework import core
from framework.backends import abstract
from framework.backends import compression

from .. import skip

# pylint: disable=no-self-use,redefined-outer-name

# Helpers


def _has_lzma():
    """Try to import backports.lzma, and return True if it exists, False if it
    doesn't.
    """
    try:
        import backports.lzma  # pylint: disable=unused-variable
    except ImportError:
        return False
    return True


def _has_xz_bin():
    """Check for an xz binary."""
    try:
        subprocess.check_call(['xz', '--help'])
    except subprocess.CalledProcessError:
        return False
    except OSError as e:  # pylint: disable=undefined-variable
        if e.errno != 2:
            raise
        return False
    return True


requires_lzma = pytest.mark.skipif(  # pylint: disable=invalid-name
    six.PY2 and not _has_lzma(),
    reason="Python 2.x requires backports.lzma to run this test.")

requires_xz_bin = pytest.mark.skipif(  # pylint: disable=invalid-name
    _has_lzma() or not _has_xz_bin(),
    reason="Python 2.x requires xz binary to run this test.")

requires_any_lzma = pytest.mark.skipif(  # pylint: disable=invalid-name
    six.PY2 and not (_has_lzma() or _has_xz_bin()),
    reason="Python 2.x requires some form of xz compression to run this test.")


@pytest.yield_fixture
def env():
    with mock.patch.dict('os.environ'):
        yield os.environ

@pytest.yield_fixture
def config():
    with mock.patch('framework.backends.compression.PIGLIT_CONFIG',
                    new_callable=core.PiglitConfig) as conf:
        conf.add_section('core')
        yield conf

@pytest.yield_fixture
def compressor():
    """Workaround for yield_fixture not working with classes."""
    class Compressor(object):
        """Fixture to control the compressor/decompressor objects."""

        def __init__(self):
            self._mock_c = mock.patch.dict(
                'framework.backends.compression.COMPRESSORS', clear=False)
            self._mock_d = mock.patch.dict(
                'framework.backends.compression.DECOMPRESSORS', clear=False)

        def __enter__(self):
            self._mock_c.start()
            self._mock_d.start()
            return self

        def __exit__(self, type_, value, traceback):
            self._mock_c.stop()
            self._mock_d.stop()

        def add(self, name, func):
            assert name not in compression.COMPRESSORS
            assert name not in compression.DECOMPRESSORS
            compression.COMPRESSORS[name] = func
            compression.DECOMPRESSORS[name] = func

        def rm(self, name):  # pylint: disable=invalid-name
            assert name in compression.COMPRESSORS
            assert name in compression.DECOMPRESSORS
            del compression.COMPRESSORS[name]
            del compression.DECOMPRESSORS[name]

    with Compressor() as c:
        yield c


# Tests


@pytest.mark.parametrize("mode", ['none', 'bz2', 'gz', requires_lzma('xz')])
def test_compress(mode, tmpdir):
    """Test that each compressor that we want works.

    These only check using modules, that means on python2 this test will skip
    unless backports.lzma is available, and it will not test the the xz binary
    path.
    """
    func = compression.COMPRESSORS[mode]
    testfile = tmpdir.join('test')
    with func(six.text_type(testfile)) as f:
        f.write('foo')


@pytest.mark.parametrize("mode", ['none', 'bz2', 'gz', requires_lzma('xz')])
def test_decompress(mode, tmpdir):
    """Test that each supported decompressor works.

    See the docstring in test_compress.
    """
    comp = compression.COMPRESSORS[mode]
    dec = compression.DECOMPRESSORS[mode]
    testfile = tmpdir.join('test')

    with comp(six.text_type(testfile)) as f:
        f.write('foo')

    with dec(six.text_type(testfile)) as f:
        actual = f.read()

    assert actual == 'foo'


@skip.posix
@skip.PY3
@requires_xz_bin
class TestXZBin(object):
    """Tests for the xz bin path on python2.x."""

    def test_compress_xz_bin(self, tmpdir):
        """Test python2 xz compression using the xz binary."""
        func = compression.COMPRESSORS['xz']
        testfile = tmpdir.join('test')
        with func(six.text_type(testfile)) as f:
            f.write('foo')

    def test_decompress_xz_bin(self, tmpdir):
        """Test python2 xz decompression using the xz binary."""
        comp = compression.COMPRESSORS['xz']
        dec = compression.DECOMPRESSORS['xz']
        testfile = tmpdir.join('test')

        with comp(six.text_type(testfile)) as f:
            f.write('foo')

        with dec(six.text_type(testfile)) as f:
            actual = f.read()

        assert actual == 'foo'


class TestGetMode(object):
    """Tests for the compression.get_mode function."""

    def test_default(self, env, config):  # pylint: disable=unused-argument
        """When neither the config file nor the environment sets a value for
        compression the default value should be used.
        """
        env.clear()

        assert compression.get_mode() == compression.DEFAULT

    def test_env(self, env, config, compressor):
        """Test that when env doesn't have a PIGLIT_COMPRESSION environment
        varaible set, but the configuraiton has a compression method set that
        it is used.
        """
        compressor.add('foo', None)
        compressor.add('bar', None)
        config.set('core', 'compression', 'foo')
        env['PIGLIT_COMPRESSION'] = 'bar'

        assert compression.get_mode() == 'bar'

    def test_piglit_conf(self, env, config, compressor):
        """Test that when env doesn't have a PIGLIT_COMPRESSION environment
        varaible set, but the configuraiton has a compression method set that
        it is used.
        """
        compressor.add('foobar', None)
        config.set('core', 'compression', 'foobar')
        env.clear()

        assert compression.get_mode() == 'foobar'


@pytest.mark.parametrize("extension", ['bz2', 'gz', requires_any_lzma('xz')])
def test_duplicate_extensions(extension, tmpdir, config):
    """Tests that exersizes a bug that caused the compressed extension to be
    duplicated in some cases.
    """
    tmpdir.chdir()
    config.set('core', 'compression', extension)
    expected = 'results.txt.' + extension

    with abstract.write_compressed(expected) as f:
        f.write('foo')

    assert expected in os.listdir('.')


@pytest.mark.parametrize("orig,new", itertools.permutations(
    ['bz2', 'gz', 'xz'], 2))
def test_changed_extension(orig, new, tmpdir, config):
    """Tests that exersizes a bug that caused two extensions to be present if
    the compression method changed.
    """
    if 'xz' in [new, orig] and six.PY2 and not (_has_lzma() or _has_xz_bin()):
        pytest.skip("There is no xz compressor available.")

    tmpdir.chdir()
    config.set('core', 'compression', six.text_type(new))

    with abstract.write_compressed('results.txt.' + orig) as f:
        f.write('foo')

    assert 'results.txt.' + new in os.listdir('.')
    assert 'results.txt.' + orig not in os.listdir('.')
    assert 'results.txt.{}.{}'.format(orig, new) not in os.listdir('.')



# encoding=utf-8
# Copyright © 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the feature summary module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
try:
    import simplejson as json
except ImportError:
    import json
try:
    import mock
except ImportError:
    from unittest import mock

import pytest
import six

from framework import grouptools
from framework import results
from framework import profile
from framework.summary import feature

from .. import utils

DATA = {
    'spec@gl-1.0': {
        'include_tests': 'gl-1.0',
        'exclude_tests': '',
        'target_rate': 50,
    },
    'spec@gl-2.0': {
        'include_tests': 'gl-2.0',
        'exclude_tests': '',
        'target_rate': 50,
    },
}


def _maketest(res):
    """Helper utility to make a test with a result."""
    t = utils.Test(['foo'])
    t.result.result = res
    return t


PROFILE = profile.TestProfile()
PROFILE.test_list = profile.TestDict()
PROFILE.test_list['spec@gl-1.0@a'] = _maketest('pass')
PROFILE.test_list['spec@gl-1.0@b'] = _maketest('warn')
PROFILE.test_list['spec@gl-1.0@c'] = _maketest('pass')
PROFILE.test_list['spec@gl-1.0@d'] = _maketest('fail')
PROFILE.test_list['spec@gl-2.0@a'] = _maketest('fail')
PROFILE.test_list['spec@gl-2.0@b'] = _maketest('crash')
PROFILE.test_list['spec@gl-2.0@c'] = _maketest('pass')
PROFILE.test_list['spec@gl-2.0@d'] = _maketest('fail')


class TestFeatResult(object):
    """Tests for the FeatResult class."""

    @pytest.fixture(scope='session')
    def feature(self, tmpdir_factory):
        p = tmpdir_factory.mktemp('feature').join('p')
        # each write to p will replace the contents, so using json.dump which
        # makes a number of small writes will fail to produce anything useful.
        p.write(json.dumps(DATA))

        result = results.TestrunResult()
        for n, s in six.iteritems(PROFILE.test_list):
            result.tests[n] = s.result
        result.options['profile'] = [None]
        result.name = 'foo'

        with mock.patch('framework.summary.feature.profile.load_test_profile',
                        mock.Mock(return_value=PROFILE)):
            return feature.FeatResults([result], six.text_type(p))

    def test_basic(self, feature):
        """The fixture works."""
        pass

    def test_features(self, feature):
        """The features set is populated."""
        assert feature.features == {'spec@gl-1.0', 'spec@gl-2.0'}

    def test_feat_fractions(self, feature):
        """feat_fraction is populated."""
        assert feature.feat_fractions == \
            {'foo': {'spec@gl-1.0': (2, 4), 'spec@gl-2.0': (1, 4)}}

    def test_feat_status(self, feature):
        """feat_status is populated."""
        assert feature.feat_status == \
            {'foo': {'spec@gl-1.0': 'pass', 'spec@gl-2.0': 'fail'}}



# coding=utf-8
# Copyright (c) 2014 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This scripts generates tests for the GLSL packing functions, such as
packSnorm2x16.

In the test templates below, observe that the GLSL function's actual output is
compared against multiple expected outputs.  Given an input and a
pack/unpackfunction, there exist multiple valid outputs because the GLSL specs
permit variation in the implementation of the function. The actual output is
dependent on the GLSL compiler's and hardware's choice of rounding mode (for
example, to even or to nearest) and handling of subnormal (also called
denormalized) floating point numbers.

"""

from __future__ import print_function, division, absolute_import
import math
import optparse
import os
import sys
from collections import namedtuple
from math import copysign, fabs, fmod, frexp, isinf, isnan, modf

from numpy import int8, int16, uint8, uint16, uint32, float32

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

# pylint: disable=bad-whitespace,line-too-long
TEMPLATE_TABLE = {
    ("const", "p", "2x16"): TEMPLATES.get_template('const_pack.shader_test.mako'),
    ("const", "p",  "4x8"): TEMPLATES.get_template('const_pack.shader_test.mako'),
    ("const", "u", "2x16"): TEMPLATES.get_template('const_unpack.shader_test.mako'),
    ("const", "u",  "4x8"): TEMPLATES.get_template('const_unpack.shader_test.mako'),
    ("vs",    "p", "2x16"): TEMPLATES.get_template('vs_pack.shader_test.mako'),
    ("vs",    "p",  "4x8"): TEMPLATES.get_template('vs_pack.shader_test.mako'),
    ("vs",    "u", "2x16"): TEMPLATES.get_template('vs_unpack.shader_test.mako'),
    ("vs",    "u",  "4x8"): TEMPLATES.get_template('vs_unpack.shader_test.mako'),
    ("fs",    "p", "2x16"): TEMPLATES.get_template('fs_pack.shader_test.mako'),
    ("fs",    "p",  "4x8"): TEMPLATES.get_template('fs_pack.shader_test.mako'),
    ("fs",    "u", "2x16"): TEMPLATES.get_template('fs_unpack.shader_test.mako'),
    ("fs",    "u",  "4x8"): TEMPLATES.get_template('fs_unpack.shader_test.mako'),
}
# pylint: enable=bad-whitespace,line-too-long

# TODO: all of the invalid names should be fixed (mostly one and two letter
# variable names), but there are lots of them, and they get in the way.
# TODO: Docstrings...
# pylint: disable=invalid-name,missing-docstring

class FuncOpts(object):  # pylint: disable=too-few-public-methods
    """Options that modify the evaluation of the GLSL pack/unpack functions.

    Given an input and a pack/unpack function, there exist multiple valid
    outputs because the GLSL specs permit variation in the implementation of
    the function. The actual output is dependent on the GLSL compiler's and
    hardware's choice of rounding mode (for example, to even or to nearest).

    This class attempts to capture the permitted variation in rounding
    behavior. To select a particular behavior, pass the appropriate enum to
    the constructor, as described below.

    Rounding mode
    -------------
    For some packing functions, the GLSL ES 3.00 specification's definition of
    the function's behavior involves round(), whose behavior at
    0.5 is an implementation detail. From section 8.3 of the spec:
        The fraction 0.5 will round in a direction chosen by the
        implementation, presumably the direction that is fastest.

    The constructor parameter 'round_mode' selects the rounding behavior.
    Valid values are:
        - ROUND_TO_EVEN
        - ROUND_TO_NEAREST
    """

    ROUND_TO_EVEN = 0
    ROUND_TO_NEAREST = 1

    def __init__(self, round_mode=ROUND_TO_EVEN):
        if round_mode == FuncOpts.ROUND_TO_EVEN:
            self.__round_func = round_to_even
        elif round_mode == FuncOpts.ROUND_TO_NEAREST:
            self.__round_func = round_to_nearest
        else:
            raise Exception('Must round to even or nearest.\n'
                            'round function: {}'.format(round_mode))

    def round(self, x):
        """Round a float according to the requested rounding mode."""
        assert any(isinstance(x, T) for T in [float, float32])

        # Drop the floating-point precision from 64 to 32 bits before
        # rounding.  The loss of precision may shift the float's fractional
        # value to 0.5, which will affect the rounding.
        x = float32(x)
        return self.__round_func(x)


def clamp(x, min_, max_):
    if x < min_:
        return min_
    elif x > max_:
        return max_
    else:
        return x


def round_to_nearest(x):
    # Get fractional and integral parts.
    (f, i) = modf(x)

    if fabs(f) < 0.5:
        return i
    else:
        return i + copysign(1.0, x)


def round_to_even(x):
    # Get fractional and integral parts.
    (f, i) = modf(x)

    if fabs(f) < 0.5:
        return i
    elif fabs(f) == 0.5:
        return i + fmod(i, 2.0)
    else:
        return i + copysign(1.0, x)


def pack_2x16(pack_1x16_func, x, y, func_opts):
    """Evaluate a GLSL pack2x16 function.

    :param pack_1x16_func: the component-wise function of the GLSL pack2x16
        function
    :param x,y: each a float32
    :return: a uint32
    """
    assert isinstance(x, float32)
    assert isinstance(y, float32)

    ux = pack_1x16_func(x, func_opts)
    uy = pack_1x16_func(y, func_opts)

    assert isinstance(ux, uint16)
    assert isinstance(uy, uint16)

    return uint32((uy << 16) | ux)


def pack_4x8(pack_1x8_func, x, y, z, w, func_opts):
    # pylint: disable=too-many-arguments
    """Evaluate a GLSL pack4x8 function.

    :param pack_1x8_func: the component-wise function of the GLSL pack4x8
        function
    :param x,y,z,w: each a float32
    :return: a uint32
    """
    assert isinstance(x, float32)
    assert isinstance(y, float32)
    assert isinstance(z, float32)
    assert isinstance(w, float32)

    ux = pack_1x8_func(x, func_opts)
    uy = pack_1x8_func(y, func_opts)
    uz = pack_1x8_func(z, func_opts)
    uw = pack_1x8_func(w, func_opts)

    assert isinstance(ux, uint8)
    assert isinstance(uy, uint8)
    assert isinstance(uz, uint8)
    assert isinstance(uw, uint8)

    return uint32((uw << 24) | (uz << 16) | (uy << 8) | ux)


def unpack_2x16(unpack_1x16_func, u, _):
    """Evaluate a GLSL unpack2x16 function.

    :param unpack_1x16_func: the component-wise function of the GLSL
        unpack2x16 function
    :param u: a uint32
    :return: a 2-tuple of float32
    """
    assert isinstance(u, uint32)

    ux = uint16(u & 0xffff)
    uy = uint16(u >> 16)

    x = unpack_1x16_func(ux)
    y = unpack_1x16_func(uy)

    assert isinstance(x, float32)
    assert isinstance(y, float32)

    return (x, y)


def unpack_4x8(unpack_1x8_func, u, _):
    """Evaluate a GLSL unpack4x8 function.

    :param unpack_1x8_func: the component-wise function of the GLSL
        unpack4x8 function
    :param u: a uint32
    :return: a 4-tuple of float32
    """
    assert isinstance(u, uint32)

    ux = uint8(u & 0xff)
    uy = uint8((u >> 8) & 0xff)
    uz = uint8((u >> 16) & 0xff)
    uw = uint8((u >> 24) & 0xff)

    x = unpack_1x8_func(ux)
    y = unpack_1x8_func(uy)
    z = unpack_1x8_func(uz)
    w = unpack_1x8_func(uw)

    assert isinstance(x, float32)
    assert isinstance(y, float32)
    assert isinstance(z, float32)
    assert isinstance(w, float32)

    return (x, y, z, w)


def pack_snorm_1x8(f32, func_opts):
    """Component-wise function of packSnorm4x8."""
    assert isinstance(f32, float32)
    return uint8(int8(func_opts.round(clamp(f32, -1.0, +1.0) * 127.0)))


def pack_snorm_1x16(f32, func_opts):
    """Component-wise function of packSnorm2x16."""
    assert isinstance(f32, float32)
    return uint16(int16(func_opts.round(clamp(f32, -1.0, +1.0) * 32767.0)))


def unpack_snorm_1x8(u8):
    """Component-wise function of unpackSnorm4x8."""
    assert isinstance(u8, uint8)
    return float32(clamp(int8(u8) / 127.0, -1.0, +1.0))


def unpack_snorm_1x16(u16):
    """Component-wise function of unpackSnorm2x16."""
    assert isinstance(u16, uint16)
    return float32(clamp(int16(u16) / 32767.0, -1.0, +1.0))


def pack_unorm_1x8(f32, func_opts):
    """Component-wise function of packUnorm4x8."""
    assert isinstance(f32, float32)
    return uint8(func_opts.round(clamp(f32, 0.0, 1.0) * 255.0))


def pack_unorm_1x16(f32, func_opts):
    """Component-wise function of packUnorm2x16."""
    assert isinstance(f32, float32)
    return uint16(func_opts.round(clamp(f32, 0.0, 1.0) * 65535.0))


def unpack_unorm_1x8(u8):
    """Component-wise function of unpackUnorm4x8."""
    assert isinstance(u8, uint8)
    return float32(u8 / 255.0)


def unpack_unorm_1x16(u16):
    """Component-wise function of unpackUnorm2x16."""
    assert isinstance(u16, uint16)
    return float32(u16 / 65535.0)


def pack_half_1x16(f32, func_opts):
    """Component-wise function of packHalf2x16."""
    assert isinstance(f32, float32)

    # The bit layout of a float16 is:
    #
    #   sign:     15
    #   exponent: 10:14
    #   mantissa: 0:9
    #
    # The sign, exponent, and mantissa determine its value by:
    #
    # if e = 0 and m = 0, then zero:       (-1)^s * 0
    # if e = 0 and m != 0, then subnormal: (-1)^s * 2^(e - 14) * m / 2^10
    # if 0 < e < 31, then normal:          (-1)^s * 2^(e - 15) * (1 + m / 2^10)
    # if e = 31 and m = 0, then inf:       (-1)^s * inf
    # if e = 31 and m != 0, then nan
    #
    # where 0 <= m < 2^10.
    #
    # Some key boundary values of float16 are:
    #
    #   min_normal16  = 2^(1 - 15) * (1 + 0 / 2^10)
    #   max_normal16  = 2^(30 - 15) * (1 + 1023 / 2^10)
    #
    # The maximum float16 step value is:
    #
    #   max_step16 = 2^5
    #
    # Observe that each of the above boundary values lies in the range of
    # normal float32 values. If we represent each of the above boundary values
    # in the form returned by frexpf() for normal float32 values, 2^E
    # * F where 0.5 <= F < 1, then:
    #
    #   min_normal16 = 2^(-13) * 0.5
    #   max_normal16 = 2^16 * 0.99951171875

    # The resultant float16's sign, exponent, and mantissa bits.
    s = 0
    e = 0
    m = 0

    # Calculate sign bit.
    # Use copysign() to handle the case where x is -0.0.
    if copysign(1.0, f32) < 0.0:
        s = 1

    # To reduce the number of cases in the if-tree below, decompose `abs(f32)`
    # rather than `f32`.
    (F, E) = frexp(fabs(f32))

    # The output of frexp falls into three classes:
    #   - If f32 is NaN, then F is NaN .
    #   - If f32 is ±inf, then F is ±inf .
    #   - If f32 is ±0.0, then F is ±0.0 .
    #   - Otherwise, f32 = 2^E * F where 0.5 <= F < 1.0 .
    #
    # Since we decomposed `abs(f32)`, we only need be concerned with the
    # positive cases.
    if isnan(F):
        # The resultant float16 is NaN.
        e = 31
        m = 1
    elif isinf(F):
        # The resultant float16 is infinite.
        e = 31
        m = 0
    elif F == 0:
        # f32 is zero, therefore the resultant float16 is zero.
        e = 0
        m = 0
    elif E < -13:
        # f32 lies in the range (0.0, min_normal16). Round f32 to a nearby
        # float16 value. The resultant float16 will be either zero, subnormal,
        # or normal.
        e = 0
        m = int(func_opts.round(2**(E + 24) * F))
    elif E <= 16:
        # f32 lies in the range [min_normal16, max_normal16 + max_step16).
        # Round f32 to a nearby float16 value. The resultant float16 will be
        # either normal or infinite.
        e = int(E + 14)
        m = int(func_opts.round(2**11 * F - 2**10))
    else:
        # f32 lies in the range [max_normal16 + max_step16, inf), which is
        # outside the range of finite float16 values. The resultant float16 is
        # infinite.
        e = 31
        m = 0

    if m == 1024:
        # f32 was rounded upwards into the range of the next exponent.  This
        # correctly handles the case where f32 should be rounded up to float16
        # infinity.
        e += 1
        m = 0

    assert s == 0 or s == 1
    assert 0 <= e and e <= 31
    assert 0 <= m and m <= 1023

    return uint16((s << 15) | (e << 10) | m)


def unpack_half_1x16(u16):
    """Component-wise function of unpackHalf2x16."""
    assert isinstance(u16, uint16)

    # The bit layout of a float16 is:
    #
    #   sign:     15
    #   exponent: 10:14
    #   mantissa: 0:9
    #
    # The sign, exponent, and mantissa determine its value by:
    #
    # if e = 0 and m = 0, then zero:       (-1)^s * 0
    # if e = 0 and m != 0, then subnormal: (-1)^s * 2^(e - 14) * m / 2^10
    # if 0 < e < 31, then normal:          (-1)^s * 2^(e - 15) * (1 + m / 2^10)
    # if e = 31 and m = 0, then inf:       (-1)^s * inf
    # if e = 31 and m != 0, then nan
    #
    # where 0 <= m < 2^10.

    s = (u16 >> 15) & 0x1
    e = (u16 >> 10) & 0x1f
    m = u16 & 0x3ff

    if s == 0:
        sign = 1.0
    else:
        sign = -1.0

    if e == 0:
        return float32(sign * 2.0**(-14) * (m / 2.0**10))
    elif 1 <= e and e <= 30:
        return float32(sign * 2.0**(e - 15.0) * (1.0 + m / 2.0**10))
    elif e == 31 and m == 0:
        return float32(sign * float32("inf"))
    elif e == 31 and m != 0:
        return float32("NaN")
    else:
        raise Exception('invalid inputs')

# ----------------------------------------------------------------------------
# Inputs for GLSL functions
# ----------------------------------------------------------------------------

# This table maps GLSL pack/unpack function names to a sequence of inputs to
# the respective component-wise function. It contains four types of mappings:
#    - name of a pack2x16 function to a sequence of float32
#    - name of a pack4x8 function to a sequence of float32
#    - name of a unpack2x16 function to a sequence of uint16
#    - name of a unpack4x8 function to a sequence of uint8
full_input_table = dict()

# This table maps each GLSL pack/unpack function name to a subset of
# ``full_input_table[name]``.
#
# To sufficiently test some functions, we must test a fairly large set of
# component-wise inputs, so large that its cartesian product explodes. The
# test such functions, we test over the cartesian product of full_input_table
# and reduced_input_table. See make_inouts_for_pack_2x16.
#
reduced_input_table = dict()


def make_inputs_for_pack_snorm_2x16():
    # The domain of packSnorm2x16 is [-inf, +inf]^2. The function clamps
    # its input into the range [-1, +1]^2.
    pos = (
        0.0,  # zero
        0.1,  # near zero
        0.9,  # slightly below the clamp boundary
        1.0,  # the clamp boundary
        1.1,  # slightly above the clamp boundary
        float("+inf")
    )
    neg = tuple(reversed(tuple(-x for x in pos)))
    return tuple(float32(x) for x in pos + neg)

full_input_table["packSnorm2x16"] = make_inputs_for_pack_snorm_2x16()
reduced_input_table["packSnorm2x16"] = None

full_input_table["packSnorm4x8"] = full_input_table["packSnorm2x16"]

# XXX: Perhaps there is a better choice of test inputs?
full_input_table["unpackSnorm2x16"] = tuple(uint16(u) for u in (
    0, 1, 2, 3,
    2**15 - 1,
    2**15,
    2**15 + 1,
    2**16 - 1  # max uint16
))

# XXX: Perhaps there is a better choice of test inputs?
full_input_table["unpackSnorm4x8"] = tuple(uint8(u) for u in (
    0, 1, 2, 3,
    2**7 - 1,
    2**7,
    2**7 + 1,
    2**8 - 1  # max uint8
))

full_input_table["packUnorm2x16"] = tuple(float32(x) for x in (
    # The domain of packUnorm2x16 is [-inf, +inf]^2. The function clamps its
    # input into the range [0, 1]^2.

    "-inf",
    -0.1,  # slightly below the inner clamp boundary
    -0.0,  # infintesimally below the inner clamp boundary
    +0.0,  # the inner clamp boundary
    +0.1,  # slightly above the inner clamp boundary
    +0.9,  # slightly below the outer clamp boundary
    +1.0,  # the outer clamp boundary
    +1.1,  # slightly above the outer clamp boundary
    "+inf"
))

reduced_input_table["packUnorm2x16"] = None

full_input_table["packUnorm4x8"] = full_input_table["packUnorm2x16"]

# XXX: Perhaps there is a better choice of test inputs?
full_input_table["unpackUnorm2x16"] = full_input_table["unpackSnorm2x16"]
full_input_table["unpackUnorm4x8"] = full_input_table["unpackSnorm4x8"]


def make_inputs_for_pack_half_2x16():
    # The domain of packHalf2x16 is ([-inf, +inf] + {NaN})^2. The function
    # does not clamp its input.
    #
    # We test both -0.0 and +0.0 in order to stress the implementation's
    # handling of zero.

    subnormal_min = 2.0**(-14) * (1.0 / 2.0**10)
    normal_min = 2.0**(-14) * (1.0 + 0.0 / 2.0**10)
    normal_max = 2.0**15 * (1.0 + 1023.0 / 2.0**10)
    min_step = 2.0**(-24)
    max_step = 2.0**5

    pos = tuple(float32(x) for x in (
        # Inputs that result in 0.0 .
        0.0,
        0.0 + 0.25 * min_step,

        # A thorny input...
        #
        # if round_to_even:
        #   f16 := 0.0
        # elif round_to_nearest:
        #    f16 := subnormal_min
        #
        0.0 + 0.50 * min_step,

        # Inputs that result in a subnormal
        # float16.
        #
        0.0 + 0.75 * min_step,
        subnormal_min + 0.00 * min_step,
        subnormal_min + 0.25 * min_step,
        subnormal_min + 0.50 * min_step,
        subnormal_min + 0.75 * min_step,
        subnormal_min + 1.00 * min_step,
        subnormal_min + 1.25 * min_step,
        subnormal_min + 1.50 * min_step,
        subnormal_min + 1.75 * min_step,
        subnormal_min + 2.00 * min_step,

        normal_min - 2.00 * min_step,
        normal_min - 1.75 * min_step,
        normal_min - 1.50 * min_step,
        normal_min - 1.25 * min_step,
        normal_min - 1.00 * min_step,
        normal_min - 0.75 * min_step,

        # Inputs that result in a normal float16.
        #
        normal_min - 0.50 * min_step,
        normal_min - 0.25 * min_step,
        normal_min + 0.00 * min_step,
        normal_min + 0.25 * min_step,
        normal_min + 0.50 * min_step,
        normal_min + 0.75 * min_step,
        normal_min + 1.00 * min_step,
        normal_min + 1.25 * min_step,
        normal_min + 1.50 * min_step,
        normal_min + 1.75 * min_step,
        normal_min + 2.00 * min_step,

        2.0 * normal_min + 0.50 * min_step,
        2.0 * normal_min + 0.75 * min_step,
        2.0 * normal_min + 1.00 * min_step,

        0.5,
        1.0,
        1.5,

        normal_max - 2.00 * max_step,
        normal_max - 1.75 * max_step,
        normal_max - 1.50 * max_step,
        normal_max - 1.25 * max_step,
        normal_max - 1.00 * max_step,
        normal_max - 0.75 * max_step,
        normal_max - 0.50 * max_step,
        normal_max - 0.25 * max_step,
        normal_max + 0.00 * max_step,
        normal_max + 0.25 * max_step,

        # Inputs that result in infinity.
        #
        normal_max + 0.50 * max_step,
        normal_max + 0.75 * max_step,
        normal_max + 1.00 * max_step,
        normal_max + 2.00 * max_step,

        "+inf"))

    neg = tuple(reversed([-x for x in pos]))
    return neg + pos

full_input_table["packHalf2x16"] = make_inputs_for_pack_half_2x16()

reduced_input_table["packHalf2x16"] = tuple(float32(x) for x in (
    "-inf",
    -2.0,
    -1.0,
    -0.0,
    +0.0,
    +1.0,
    +2.0,
    "+inf"
))


def make_inputs_for_unpack_half_2x16():
    # For each of the two classes of float16 values, subnormal and normalized,
    # below are listed the exponent and mantissa of the class's boundary
    # values and some values slightly inside the bounds.
    # pylint: disable=bad-whitespace
    bounds = (
        (0,     0),  # zero
        (0,     1),  # subnormal_min
        (0,     2),  # subnormal_min + min_step
        (0,  1022),  # subnormal_max - min_step
        (0,  1023),  # subnormal_max
        (1,     0),  # normal_min
        (1,     1),  # normal_min + min_step
        (30, 1022),  # normal_max - max_step
        (30, 1023),  # normal_max
        (31,    0)   # inf
    )
    # pylint: enable=bad-whitespace

    def make_uint16(s, e, m):
        return uint16((s << 15) | (e << 10) | m)

    pos = tuple(make_uint16(0, e, m) for (e, m) in bounds)
    neg = tuple(make_uint16(1, e, m) for (e, m) in reversed(bounds))
    return neg + pos

full_input_table["unpackHalf2x16"] = make_inputs_for_unpack_half_2x16()

# ----------------------------------------------------------------------------
# Expected outputs for GLSL functions
# ----------------------------------------------------------------------------

# For a given input to a GLSL function, InOutTuple lists all valid outputs.
#
# There are multiple types of InOutTuple, described below. In each
# description, the numerical types actually refer to strings that represent
# a GLSL literal of that type.
#
#   - That for a pack2x16 function: the input is a 2-tuple of float32 and each
#     output is a uint32. For example, ``InOutTuple(input=("0.0", "0.0"),
#     valid_outputs=("0u", "0u", "0u"))``.
#
#   - That for a unpack2x16 function: the input is a uint32 and each output is
#     a 2-tuple of float32. For example, ``InOutTuple(input="0x80000000u",
#     valid_outputs=(("0.0", "-0.0"),))``.
#
InOutTuple = namedtuple("InOutTuple", ("input", "valid_outputs"))


def glsl_literal(x):
    """Convert the given number to a string that represents a GLSL literal.

    :param x: a uint32 or float32
    """
    if isinstance(x, uint32):
        return "{0}u".format(uint32(x))
    elif isinstance(x, float32):
        if math.isnan(x):
            # GLSL ES 3.00 and GLSL 4.10 do not require implementations to
            # support NaN, so we do not test it.
            raise Exception('NaN is not tested.')
        elif math.isinf(x):
            # GLSL ES 3.00 lacks a literal for infinity. However, ±1.0e256
            # suffices because it lies sufficientlyoutside the range of finite
            # float32 values.
            #
            #  From page 31 of the GLSL ES 3.00 spec:
            #
            #   If the value of the floating point number is too large (small)
            #   to be stored as a single precision value, it is converted to
            #   positive (negative) infinity.
            #
            return repr(copysign(1.0e256, x))
        elif x == 0 and copysign(1.0, x) == -1.0:
            # Workaround for numpy-1.7.0, in which repr(float32(-0.0)) does
            # not return a float literal.
            # See https://github.com/numpy/numpy/issues/2935 .
            return "-0.0"
        else:
            return repr(x)
    else:
        raise Exception('Unsupported GLSL litteral')


def make_inouts_for_pack_2x16(pack_1x16_func,
                              all_float32_inputs,
                              reduced_inputs=None):
    """Determine valid outputs for a given GLSL pack2x16 function.

    If the reduced_float32_inputs parameter is None, then it is assumed to be
    the same as all_float32_inputs.

    The set of vec2 inputs constructed by this function is the union of
    cartesian products:
      (all_float32_inputs x reduced_inputs)
      + (reduced_inputs x all_float32_inputs)

    :param pack_1x16_func: the component-wise function of the pack2x16
        function
    :param float32_inputs: a sequence of inputs to pack_1x16_func
    :return: a sequence of InOutTuple
    """
    inout_seq = []

    func_opt_seq = (FuncOpts(FuncOpts.ROUND_TO_EVEN),
                    FuncOpts(FuncOpts.ROUND_TO_NEAREST))

    if reduced_inputs is None:
        reduced_inputs = all_float32_inputs

    def add_vec2_input(x, y):
        assert isinstance(x, float32)
        assert isinstance(y, float32)

        valid_outputs = []
        for func_opts in func_opt_seq:
            u32 = pack_2x16(pack_1x16_func, x, y, func_opts)
            assert isinstance(u32, uint32)
            valid_outputs.append(glsl_literal(u32))

        inout_seq.append(
            InOutTuple(input=(glsl_literal(x), glsl_literal(y)),
                       valid_outputs=valid_outputs))

    for y in reduced_inputs:
        for x in all_float32_inputs:
            add_vec2_input(x, y)
            add_vec2_input(y, x)

    return inout_seq


def make_inouts_for_pack_4x8(pack_1x8_func, float32_inputs):
    """Determine valid outputs for a given GLSL pack4x8 function.

    :param pack_1x8_func: the component-wise function of the pack4x8
        function
    :param float32_inputs: a sequence of inputs to pack_1x8_func
    :return: a sequence of InOutTuple
    """
    inout_seq = []

    func_opt_seq = (FuncOpts(FuncOpts.ROUND_TO_EVEN),
                    FuncOpts(FuncOpts.ROUND_TO_NEAREST))

    for y in float32_inputs:
        for x in float32_inputs:
            assert isinstance(x, float32)

            valid_outputs_0 = []
            valid_outputs_1 = []
            for func_opts in func_opt_seq:
                u32_0 = pack_4x8(pack_1x8_func, x, y, x, y, func_opts)
                u32_1 = pack_4x8(pack_1x8_func, x, x, y, y, func_opts)
                assert isinstance(u32_0, uint32)
                assert isinstance(u32_1, uint32)
                valid_outputs_0.append(glsl_literal(u32_0))
                valid_outputs_1.append(glsl_literal(u32_1))

            inout_seq.append(
                InOutTuple(input=(glsl_literal(x), glsl_literal(y),
                                  glsl_literal(x), glsl_literal(y)),
                           valid_outputs=valid_outputs_0))
            inout_seq.append(
                InOutTuple(input=(glsl_literal(x), glsl_literal(x),
                                  glsl_literal(y), glsl_literal(y)),
                           valid_outputs=valid_outputs_1))
    return inout_seq


def make_inouts_for_unpack_2x16(unpack_1x16_func, uint16_inputs):
    """Determine expected outputs of a given GLSL unpack2x16 function.

    :param unpack_1x16_func: the component-wise function of the unpack2x16
        function
    :param uint16_inputs: a sequence of inputs to unpack_1x16_func
    :return: a sequence of InOutTuple
    """
    inout_seq = []
    func_opts = FuncOpts()

    for y in uint16_inputs:
        for x in uint16_inputs:
            assert isinstance(x, uint16)
            u32 = uint32((y << 16) | x)
            vec2 = unpack_2x16(unpack_1x16_func, u32, func_opts)
            assert isinstance(vec2[0], float32)
            assert isinstance(vec2[1], float32)
            inout_seq.append(
                InOutTuple(input=glsl_literal(u32),
                           valid_outputs=[(glsl_literal(vec2[0]),
                                           glsl_literal(vec2[1]))]))

    return inout_seq


def make_inouts_for_unpack_4x8(unpack_1x8_func, uint8_inputs):
    """Determine expected outputs of a given GLSL unpack4x8 function.

    :param unpack_1x8_func: the component-wise function of the unpack4x8
        function
    :param uint8_inputs: a sequence of inputs to unpack_1x8_func
    :return: a sequence of InOutTuple
    """
    inout_seq = []

    func_opts = FuncOpts()

    for y in uint8_inputs:
        for x in uint8_inputs:
            assert isinstance(x, uint8)
            u32_0 = uint32((y << 24) | (x << 16) | (y << 8) | x)
            u32_1 = uint32((y << 24) | (y << 16) | (x << 8) | x)

            valid_outputs_0 = []
            valid_outputs_1 = []
            vec4_0 = unpack_4x8(unpack_1x8_func, u32_0, func_opts)
            vec4_1 = unpack_4x8(unpack_1x8_func, u32_1, func_opts)
            assert isinstance(vec4_0[0], float32)
            assert isinstance(vec4_0[1], float32)
            assert isinstance(vec4_0[2], float32)
            assert isinstance(vec4_0[3], float32)
            assert isinstance(vec4_1[0], float32)
            assert isinstance(vec4_1[1], float32)
            assert isinstance(vec4_1[2], float32)
            assert isinstance(vec4_1[3], float32)
            valid_outputs_0.append((glsl_literal(vec4_0[0]),
                                    glsl_literal(vec4_0[1]),
                                    glsl_literal(vec4_0[2]),
                                    glsl_literal(vec4_0[3])))
            valid_outputs_1.append((glsl_literal(vec4_1[0]),
                                    glsl_literal(vec4_1[1]),
                                    glsl_literal(vec4_1[2]),
                                    glsl_literal(vec4_1[3])))

            inout_seq.append(InOutTuple(input=glsl_literal(u32_0),
                                        valid_outputs=valid_outputs_0))
            inout_seq.append(InOutTuple(input=glsl_literal(u32_1),
                                        valid_outputs=valid_outputs_1))

    return inout_seq

# This table maps GLSL pack/unpack function names to the precision of their
# return type.
result_precision_table = {
    "packSnorm2x16": "highp",
    "packSnorm4x8": "highp",
    "packUnorm2x16": "highp",
    "packUnorm4x8": "highp",
    "packHalf2x16":  "highp",
    "unpackSnorm2x16": "highp",
    "unpackSnorm4x8": "highp",
    "unpackUnorm2x16": "highp",
    "unpackUnorm4x8": "highp",
    "unpackHalf2x16":  "mediump"
}

# This table maps GLSL pack/unpack function names to a sequence of InOutTuple.
inout_table = {
    "packSnorm2x16": make_inouts_for_pack_2x16(
        pack_snorm_1x16, full_input_table["packSnorm2x16"],
        reduced_input_table["packSnorm2x16"]),
    "packSnorm4x8": make_inouts_for_pack_4x8(
        pack_snorm_1x8, full_input_table["packSnorm4x8"]),
    "packUnorm2x16": make_inouts_for_pack_2x16(
        pack_unorm_1x16, full_input_table["packUnorm2x16"],
        reduced_input_table["packUnorm2x16"]),
    "packUnorm4x8": make_inouts_for_pack_4x8(
        pack_unorm_1x8, full_input_table["packUnorm4x8"]),
    "packHalf2x16":  make_inouts_for_pack_2x16(
        pack_half_1x16, full_input_table["packHalf2x16"],
        reduced_input_table["packHalf2x16"]),
    "unpackSnorm2x16": make_inouts_for_unpack_2x16(
        unpack_snorm_1x16, full_input_table["unpackSnorm2x16"]),
    "unpackSnorm4x8": make_inouts_for_unpack_4x8(
        unpack_snorm_1x8, full_input_table["unpackSnorm4x8"]),
    "unpackUnorm2x16": make_inouts_for_unpack_2x16(
        unpack_unorm_1x16, full_input_table["unpackUnorm2x16"]),
    "unpackUnorm4x8": make_inouts_for_unpack_4x8(
        unpack_unorm_1x8, full_input_table["unpackUnorm4x8"]),
    "unpackHalf2x16": make_inouts_for_unpack_2x16(
        unpack_half_1x16, full_input_table["unpackHalf2x16"])
}


# ----------------------------------------------------------------------------
# Generate test files
# ----------------------------------------------------------------------------


FuncInfo = namedtuple('FuncInfo', ['name', 'dimension', 'result_precision',
                                   'inout_seq', 'num_valid_outputs',
                                   'vector_type', 'requirements', 'exact'])

def func_info(name, requirements):
    """Factory function for information for a GLSL pack/unpack function.

    Properties
    ----------
    - name: Name of the GLSL function, such as "packSnorm2x16".

    - dimension: Dimension of the GLSL function, such as "2x16".

    - result_precision: Precision of the GLSL function's return type, such as
      "highp".

    - inout_seq: A sequence of InOutTuple.  The generated test file will test
      all inputs listed in the sequence.

    - num_valid_outputs: The number of valid outputs for each input of
      self.inout_seq. (We assume that each input has the  same number of valid
      outputs).

    - vector_type: The type of the GLSL function's parameter  or return value.
      E.g., vec4 for a 4x8 function and vec2 for a 2x16 function.

    - requirements: A set of API/extension requirments to be listed in the
      .shader_test's [requires] section.

    - exact: Whether the generated results must be exact (e.g., 0.0 and 1.0
      should always be converted exactly).

    """

    if name.endswith("2x16"):
        dimension = "2x16"
        vector_type = "vec2"
    elif name.endswith("4x8"):
        dimension = "4x8"
        vector_type = "vec4"
    else:
        raise Exception('Invalid pack type {}'.format(name))

    inout_seq = inout_table[name]

    return FuncInfo(name, dimension, result_precision_table[name],
                    inout_seq, len(inout_seq[0].valid_outputs), vector_type,
                    requirements, name.endswith("unpackHalf2x16"))


class ShaderTest(object):
    """A .shader_test file."""

    @staticmethod
    def all_tests():
        requirements = "GLSL >= 1.30\nGL_ARB_shading_language_packing"
        ARB_shading_language_packing_funcs = (
            func_info("packSnorm2x16", requirements),
            func_info("packSnorm4x8", requirements),
            func_info("packUnorm2x16", requirements),
            func_info("packUnorm4x8", requirements),
            func_info("packHalf2x16", requirements),
            func_info("unpackSnorm2x16", requirements),
            func_info("unpackSnorm4x8", requirements),
            func_info("unpackUnorm2x16", requirements),
            func_info("unpackUnorm4x8", requirements),
            func_info("unpackHalf2x16", requirements)
            )

        requirements = "GL ES >= 3.0\nGLSL ES >= 3.00"
        glsl_es_300_funcs = (
            func_info("packSnorm2x16", requirements),
            func_info("packUnorm2x16", requirements),
            func_info("packHalf2x16", requirements),
            func_info("unpackSnorm2x16", requirements),
            func_info("unpackUnorm2x16", requirements),
            func_info("unpackHalf2x16", requirements)
            )

        execution_stages = ("const", "vs", "fs")

        for s in execution_stages:
            for f in glsl_es_300_funcs:
                yield ShaderTest(f, s, "glsl-es-3.00")
            for f in ARB_shading_language_packing_funcs:
                yield ShaderTest(f, s, "ARB_shading_language_packing")

    def __init__(self, funcinfo, execution_stage, api):
        assert isinstance(funcinfo, FuncInfo)
        assert execution_stage in ("const", "vs", "fs")
        assert api in ("glsl-es-3.00", "ARB_shading_language_packing")

        self.__template = TEMPLATE_TABLE[(execution_stage,
                                          funcinfo.name[0],
                                          funcinfo.dimension)]
        self.__func_info = funcinfo
        self.__filename = os.path.join(
            "spec",
            api.lower(),
            "execution",
            "built-in-functions",
            "{0}-{1}.shader_test".format(execution_stage, funcinfo.name))

    @property
    def filename(self):
        return self.__filename

    def write_file(self):
        dirname = os.path.dirname(self.filename)
        utils.safe_makedirs(dirname)

        with open(self.filename, "w") as f:
            f.write(self.__template.render_unicode(func=self.__func_info))


def main():
    parser = optparse.OptionParser(
        description="Generate shader tests that test the built-inpacking "
                    "functions",
        usage="usage: %prog [-h] [--names-only]")
    parser.add_option(
        '--names-only',
        dest='names_only',
        action='store_true',
        help="Don't output files, just generate a list of filenames to stdout")

    (options, args) = parser.parse_args()

    if len(args) != 0:
        # User gave extra args.
        parser.print_help()
        sys.exit(1)

    for test in ShaderTest.all_tests():
        print(test.filename)

        # Some test files take a long time to generate, so provide status
        # updates to the user immediately.
        sys.stdout.flush()

        if not options.names_only:
            test.write_file()

if __name__ == '__main__':
    main()



# coding=utf-8
"""A profile that runs only GLSLParserTest instances."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os

from framework import grouptools
from framework.profile import TestProfile
from framework.test.glsl_parser_test import GLSLParserTest, GLSLParserNoConfigError
from framework.test.piglit_test import ASMParserTest, ROOT_DIR
from .py_modules.constants import GENERATED_TESTS_DIR, TESTS_DIR

__all__ = ['profile']

profile = TestProfile()

# Find and add all shader tests.
basepath = os.path.normpath(os.path.join(TESTS_DIR, '..'))
gen_basepath = os.path.relpath(os.path.join(GENERATED_TESTS_DIR, '..'), basepath)

for basedir in [TESTS_DIR, GENERATED_TESTS_DIR]:
    isgenerated = basedir == GENERATED_TESTS_DIR
    for dirpath, _, filenames in os.walk(basedir):
        groupname = grouptools.from_path(os.path.relpath(dirpath, basedir))
        for filename in filenames:
            testname, ext = os.path.splitext(filename)
            if ext in ['.vert', '.tesc', '.tese', '.geom', '.frag', '.comp']:
                dirname = os.path.relpath(dirpath, basepath)
                filepath = os.path.join(dirname, filename)
                if isgenerated:
                    installpath = os.path.relpath(filepath, gen_basepath)
                else:
                    installpath = None

                try:
                    test = GLSLParserTest.new(filepath, installpath)
                except GLSLParserNoConfigError:
                    # In the event that there is no config assume that it is a
                    # legacy test, and continue
                    continue

                # For glslparser tests you can have multiple tests with the
                # same name, but a different stage, so keep the extension.
                testname = filename
            else:
                continue

            group = grouptools.join(groupname, testname)
            assert group not in profile.test_list, group

            profile.test_list[group] = test

# Collect and add all asmparsertests
for basedir in [TESTS_DIR, GENERATED_TESTS_DIR]:
    _basedir = os.path.join(basedir, 'asmparsertest', 'shaders')
    for dirpath, _, filenames in os.walk(_basedir):
        base_group = grouptools.from_path(os.path.join(
            'asmparsertest', os.path.relpath(dirpath, _basedir)))
        type_ = os.path.basename(dirpath)

        dirname = os.path.relpath(dirpath, os.path.join(basedir, '..'))
        for filename in filenames:
            if not os.path.splitext(filename)[1] == '.txt':
                continue

            group = grouptools.join(base_group, filename)
            profile.test_list[group] = ASMParserTest(
                type_, os.path.join(dirname, filename))



# coding=utf-8
# Copyright (c) 2014 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module implementing tests for test/all.py and friends

This module only provides tests for native piiglit tests (OpenGL and OpenCL),
it does not provide tests for non native tests that use piglit (oglconfrom, etc)

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import importlib
import os.path

import pytest

MODULES = [
    'all',
    pytest.mark.skipif(not os.path.exists('generated_tests/cl/builtin'),
                       reason='Requires that CL functionality is built')('cl'),
    'cpu',
    'glslparser',
    'gpu',
    'llvmpipe',
    'quick',
    'shader',
]


@pytest.mark.parametrize('name', MODULES)
@pytest.mark.skipif(not os.path.exists('bin'),
                    reason='This test requires the C layer to be built.')
def test_import(name):
    """Test that each built-in module can be imported."""

    importlib.import_module('tests.{}'.format(name))



# coding=utf-8
# Copyright (c) 2014 Intel Corporation
# Copyright (c) 2018 Advanced Micro Devices, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Generate spec/EXT_gpu_shader4 tests """

from __future__ import print_function, division, absolute_import
import os
import collections

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

vec = ['vec2', 'vec3', 'vec4']
ivec = ['ivec2', 'ivec3', 'ivec4']
uvec = ['uvec2', 'uvec3', 'uvec4']

genType = ['float'] + vec
genIType = ['int'] + ivec
genUType = ['unsigned int'] + uvec

ShiftOpTypes = [
    ('int', 'int'), ('unsigned int', 'unsigned int'),
    ('ivec2', 'ivec2'), ('ivec3', 'ivec3'), ('ivec4', 'ivec4'),
    ('uvec2', 'uvec2'), ('uvec3', 'uvec3'), ('uvec4', 'uvec4'),

    ('ivec2', 'int'), ('ivec3', 'int'), ('ivec4', 'int'),
    ('uvec2', 'unsigned int'), ('uvec3', 'unsigned int'), ('uvec4', 'unsigned int'),
]

BinaryOpTypes = ShiftOpTypes + [
    ('int', 'ivec2'), ('int', 'ivec3'), ('int', 'ivec4'),
    ('unsigned int', 'uvec2'), ('unsigned int', 'uvec3'), ('unsigned int', 'uvec4'),
]

OperatorParams = collections.namedtuple(
    'OperatorParams', ['name', 'op', 'types'])

UNARY_OP_TESTS = [
    OperatorParams('bitwise-not', '~', genIType + genUType),
]

BINARY_OP_TESTS = [
    OperatorParams('mod', '%', BinaryOpTypes),
    OperatorParams('bitwise-and', '&', BinaryOpTypes),
    OperatorParams('bitwise-or', '|', BinaryOpTypes),
    OperatorParams('xor', '^', BinaryOpTypes),
    OperatorParams('lshift', '<<', ShiftOpTypes),
    OperatorParams('rshift', '>>', ShiftOpTypes),
]


Type1Params = collections.namedtuple(
    'Type1Params', ['func', 'type1'])
Type2Params = collections.namedtuple(
    'Type2Params', ['func', 'type1', 'type2'])

UNARY_TESTS = [
    Type1Params('abs', genIType),
    Type1Params('sign', genIType),
    Type1Params('truncate', genType),
    Type1Params('round', genType),
]

BINARY_TESTS = [
    Type1Params('min', genIType),
    Type2Params('min', ivec, 'int'),
    Type1Params('min', genUType),
    Type2Params('min', uvec, 'unsigned int'),

    Type1Params('max', genIType),
    Type2Params('max', ivec, 'int'),
    Type1Params('max', genUType),
    Type2Params('max', uvec, 'unsigned int'),
]

TERNARY_TESTS = [
    Type1Params('clamp', genIType),
    Type2Params('clamp', ivec, 'int'),
    Type1Params('clamp', genUType),
    Type2Params('clamp', uvec, 'unsigned int'),
]

VEC_COMPARE_TESTS = [
    Type1Params('lessThan', uvec),
    Type1Params('lessThanEqual', uvec),
    Type1Params('greaterThan', uvec),
    Type1Params('greaterThanEqual', uvec),
    Type1Params('equal', uvec),
    Type1Params('notEqual', uvec),
]


FLOAT = ['']
INT = ['i', 'u']
ALL = FLOAT + INT

FetchParams = collections.namedtuple(
    'FetchParams', ['sampler', 'coord', 'offsetcoord', 'variants'])

FETCH_TESTS = [
    FetchParams('1D', 'int', 'int', ALL),
    FetchParams('2D', 'ivec2', 'ivec2', ALL),
    FetchParams('3D', 'ivec3', 'ivec3', ALL),
    FetchParams('2DRect', 'ivec2', 'ivec2', ALL),
    FetchParams('1DArray', 'ivec2', 'int', ALL),
    FetchParams('2DArray', 'ivec3', 'ivec2', ALL),
    FetchParams('Buffer', 'int', '', ALL),
]

SIZE_TESTS = FETCH_TESTS + [
    FetchParams('Cube', 'ivec2', '', ALL),
]


TextureParams = collections.namedtuple(
    'TextureParams', ['func', 'sampler', 'coord', 'offsetcoord', 'variants'])

# Variants: std, bias, lod
COMMON_TEXTURE_TESTS = [
    # Inherited from GLSL 1.10, test integer textures
    TextureParams('texture1D', '1D', 'float', 'int', INT),
    TextureParams('texture1DProj', '1D', 'vec2', 'int', INT),
    TextureParams('texture1DProj', '1D', 'vec4', 'int', INT),

    TextureParams('texture2D', '2D', 'vec2', 'ivec2', INT),
    TextureParams('texture2DProj', '2D', 'vec3', 'ivec2', INT),
    TextureParams('texture2DProj', '2D', 'vec4', 'ivec2', INT),

    TextureParams('texture3D', '3D', 'vec3', 'ivec3', INT),
    TextureParams('texture3DProj', '3D', 'vec4', 'ivec3', INT),

    TextureParams('textureCube', 'Cube', 'vec3', '', INT),

    # Added by GL_EXT_gpu_shader4
    TextureParams('texture1DArray', '1DArray', 'vec2', 'int', ALL),
    TextureParams('texture2DArray', '2DArray', 'vec3', 'ivec2', ALL),

    TextureParams('shadow1D', '1DShadow', 'vec3', 'int', FLOAT),
    TextureParams('shadow2D', '2DShadow', 'vec3', 'ivec2', FLOAT),
    TextureParams('shadow1DProj', '1DShadow', 'vec4', 'int', FLOAT),
    TextureParams('shadow2DProj', '2DShadow', 'vec4', 'ivec2', FLOAT),
    TextureParams('shadow1DArray', '1DArrayShadow', 'vec3', 'int', FLOAT),
]

TEXTURE_TESTS = COMMON_TEXTURE_TESTS + [
    TextureParams('shadow2DArray', '2DArrayShadow', 'vec4', 'ivec2', FLOAT),
    TextureParams('shadowCube', 'CubeShadow', 'vec4', '', FLOAT),

    # Inherited from ARB_texture_rectangle, test integer textures
    TextureParams('texture2DRect', '2DRect', 'vec2', 'ivec2', INT),
    TextureParams('texture2DRectProj', '2DRect', 'vec3', 'ivec2', INT),
    TextureParams('texture2DRectProj', '2DRect', 'vec4', 'ivec2', INT),
    TextureParams('shadow2DRect', '2DRectShadow', 'vec3', 'ivec2', FLOAT),
    TextureParams('shadow2DRectProj', '2DRectShadow', 'vec4', 'ivec2', FLOAT),
]

LOD_TESTS = COMMON_TEXTURE_TESTS
BIAS_TESTS = COMMON_TEXTURE_TESTS


GradParams = collections.namedtuple(
    'GradParams', ['coord', 'grad', 'offsetcoord', 'sampler', 'func', 'variants'])

GRAD_TESTS = [
    GradParams('float', 'float', 'int', '1D', 'texture1D', ALL),
    GradParams('vec2', 'float', 'int', '1D', 'texture1DProj', ALL),
    GradParams('vec4', 'float', 'int', '1D', 'texture1DProj', ALL),
    GradParams('vec2', 'float', 'int', '1DArray', 'texture1DArray', ALL),

    GradParams('vec2', 'vec2', 'ivec2', '2D', 'texture2D', ALL),
    GradParams('vec3', 'vec2', 'ivec2', '2D', 'texture2DProj', ALL),
    GradParams('vec4', 'vec2', 'ivec2', '2D', 'texture2DProj', ALL),
    GradParams('vec3', 'vec2', 'ivec2', '2DArray', 'texture2DArray', ALL),

    GradParams('vec3', 'vec3', 'ivec3', '3D', 'texture3D', ALL),
    GradParams('vec4', 'vec3', 'ivec3', '3D', 'texture3DProj', ALL),

    GradParams('vec3', 'vec3', '', 'Cube', 'textureCube', ALL),

    GradParams('vec3', 'float', 'int', '1DShadow', 'shadow1D', FLOAT),
    GradParams('vec4', 'float', 'int', '1DShadow', 'shadow1DProj', FLOAT),
    GradParams('vec3', 'float', 'int', '1DArrayShadow', 'shadow1DArray', FLOAT),

    GradParams('vec3', 'vec2', 'ivec2', '2DShadow', 'shadow2D', FLOAT),
    GradParams('vec4', 'vec2', 'ivec2', '2DShadow', 'shadow2DProj', FLOAT),
    GradParams('vec4', 'vec2', 'ivec2', '2DArrayShadow', 'shadow2DArray', FLOAT),

    GradParams('vec2', 'vec2', 'ivec2', '2DRect', 'texture2DRect', ALL),
    GradParams('vec3', 'vec2', 'ivec2', '2DRect', 'texture2DRectProj', ALL),
    GradParams('vec4', 'vec2', 'ivec2', '2DRect', 'texture2DRectProj', ALL),

    GradParams('vec3', 'vec2', 'ivec2', '2DRectShadow', 'shadow2DRect', FLOAT),
    GradParams('vec4', 'vec2', 'ivec2', '2DRectShadow', 'shadow2DRectProj', FLOAT),

    GradParams('vec4', 'vec3', '', 'CubeShadow', 'shadowCube', FLOAT),
]

def get_swizzle(vtype):
    """
    Expand the type to vec4.
    """
    if '3' in vtype:
        return '.xyzz';
    if '2' in vtype:
        return '.xyyy';
    return '';

def get_bvec(vtype):
    """
    Return the corresponding bvec type.
    """
    if '4' in vtype:
        return 'bvec4';
    if '3' in vtype:
        return 'bvec3';
    if '2' in vtype:
        return 'bvec2';
    return 'invalid';

def get_extensions(sampler):
    """ If this test uses GL_ARB_texture_rectangle add it

    GL_ARB_texture_rectangle is an odd extension, it is on by default, so don't
    generate a #extension in the shader, just in the config block.

    """
    if 'Rect' in sampler:
        return 'GL_EXT_gpu_shader4 GL_ARB_texture_rectangle'

    if 'Array' in sampler:
        return 'GL_EXT_gpu_shader4 GL_EXT_texture_array'

    if 'Buffer' in sampler:
        return 'GL_EXT_gpu_shader4 GL_EXT_texture_buffer_object'

    return 'GL_EXT_gpu_shader4'


def main():
    """ Main function

    Writes tests to generated_tests/spec/ext_gpu_shader4/ directory

    """
    dirname = 'spec/ext_gpu_shader4/compiler'
    utils.safe_makedirs(dirname)

    for params in UNARY_OP_TESTS:
        for type1 in params.types:
            name = os.path.join(
                dirname,
                "{func}-{type1}".format(
                    func=params.name,
                    type1=type1.replace(' ', '_')))

            for stage in ['frag', 'vert']:
                filename = '{0}.{1}'.format(name, stage)
                print(filename)
                with open(filename, 'w+') as f:
                    f.write(TEMPLATES.get_template(
                        '{0}.{1}.mako'.format('unary_op', stage)).render_unicode(
                            param=params,
                            type1=type1,
                            swizzle=get_swizzle(type1)))

    for params in BINARY_OP_TESTS:
        for types in params.types:
            name = os.path.join(
                dirname,
                "{func}-{type1}-{type2}".format(
                    func=params.name,
                    type1=types[0].replace(' ', '_'),
                    type2=types[1].replace(' ', '_')))
            result_type = types[0] if 'int' in types[1] else types[1]

            for stage in ['frag', 'vert']:
                filename = '{0}.{1}'.format(name, stage)
                print(filename)
                with open(filename, 'w+') as f:
                    f.write(TEMPLATES.get_template(
                        '{0}.{1}.mako'.format('binary_op', stage)).render_unicode(
                            param=params,
                            types=types,
                            result_type=result_type,
                            swizzle=get_swizzle(result_type)))

    for tests in [('unary', UNARY_TESTS),
                  ('binary', BINARY_TESTS),
                  ('ternary', TERNARY_TESTS),
                  ('vec_compare', VEC_COMPARE_TESTS)]:
        for params in tests[1]:
            for type1 in params.type1:
                name = os.path.join(
                    dirname,
                    "{func}-{type1}{type2}".format(
                        func=params.func,
                        type1=type1.replace(' ', '_'),
                        type2=('-' + params.type2 if 'type2' in params._fields else '').replace(' ', '_')))

                for stage in ['frag', 'vert']:
                    filename = '{0}.{1}'.format(name, stage)
                    print(filename)
                    with open(filename, 'w+') as f:
                        f.write(TEMPLATES.get_template(
                            '{0}.{1}.mako'.format(tests[0], stage)).render_unicode(
                                param=params,
                                type1=type1,
                                type2=params.type2 if 'type2' in params._fields else type1,
                                swizzle=get_swizzle(type1),
                                bvec=get_bvec(type1)))

    for tests in [('texture_size', SIZE_TESTS, ''),
                  ('texel_fetch', FETCH_TESTS, ''),
                  ('tex', TEXTURE_TESTS, ''),
                  ('tex_bias', BIAS_TESTS, ''),
                  ('tex_lod', LOD_TESTS, ''),
                  ('tex_grad', GRAD_TESTS, ''),
                  ('texel_fetch', FETCH_TESTS, 'Offset'),
                  ('tex', TEXTURE_TESTS, 'Offset'),
                  ('tex_bias', BIAS_TESTS, 'Offset'),
                  ('tex_lod', LOD_TESTS, 'Offset'),
                  ('tex_grad', GRAD_TESTS, 'Offset')]:
        for params in tests[1]:
            if tests[2] == 'Offset' and params.offsetcoord == '':
                continue

            for prefix in params.variants:
                name = os.path.join(
                    dirname,
                    "{0}{func}{offset}-{prefix}sampler{sampler}-{coord}".format(
                        tests[0],
                        func='-' + params.func if 'func' in params._fields else '',
                        offset=tests[2],
                        prefix=prefix,
                        sampler=params.sampler,
                        coord=params.coord))

                stages = ['frag']
                if tests[0] != 'tex_bias':
                    stages = stages + ['vert']

                for stage in stages:
                    filename = '{0}.{1}'.format(name, stage)
                    print(filename)
                    with open(filename, 'w+') as f:
                        f.write(TEMPLATES.get_template(
                            '{0}.{1}.mako'.format(tests[0], stage)).render_unicode(
                                param=params,
                                extensions=get_extensions(params.sampler),
                                prefix=prefix,
                                swizzle=get_swizzle(params.coord),
                                offset=tests[2]))

if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2014-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the piglit_test module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import textwrap
try:
    from unittest import mock
except ImportError:
    import mock

import pytest


from framework import status
from framework.options import _Options as Options
from framework.test.base import TestIsSkip as _TestIsSkip
from framework.test.piglit_test import PiglitBaseTest, PiglitGLTest

# pylint: disable=no-self-use
# pylint: disable=protected-access


class TestPiglitBaseTest(object):
    """Tests for the PiglitBaseTest class."""

    class TestIntepretResult(object):
        """Tests for PiglitBaseTest.interpret_results."""

        def test_basic(self):
            """A basic sanity test with nothing tricky."""
            test = PiglitBaseTest(['foo'])
            test.result.out = 'PIGLIT: {"result": "pass"}\n'
            test.result.returncode = 0
            test.interpret_result()
            assert test.result.result is status.PASS

        def test_stdout(self):
            """Separates the actual stdout printing from the PIGLIT protocol.
            """
            test = PiglitBaseTest(['foo'])
            test.result.out = textwrap.dedent("""\
                This is output

                more output
                PIGLIT: {"result": "pass"}
                and stuff""")
            test.result.returncode = 0
            test.interpret_result()

            assert test.result.result is status.PASS
            assert test.result.out == textwrap.dedent("""\
                This is output

                more output
                and stuff""")

        def test_with_subtests(self):
            """works with one subtest."""
            test = PiglitBaseTest(['foo'])
            test.result.out = textwrap.dedent("""\
                PIGLIT: {"result": "pass"}
                PIGLIT: {"subtest": {"subtest": "pass"}}""")
            test.result.returncode = 0
            test.interpret_result()
            assert test.result.subtests['subtest'] is status.PASS

        def test_with_multiple_subtests(self):
            """Works with multiple subtests.

            Including that it doesn't only take the last subtest, but counts
            all of them.
            """
            test = PiglitBaseTest(['a', 'command'])
            test.result.out = textwrap.dedent("""\
                PIGLIT: {"result": "pass"}
                PIGLIT: {"subtest": {"test1": "pass"}}
                PIGLIT: {"subtest": {"test2": "pass"}}""")
            test.result.returncode = 0
            test.interpret_result()

            assert dict(test.result.subtests) == \
                {'test1': 'pass', 'test2': 'pass'}


class TestPiglitGLTest(object):
    """tests for the PiglitGLTest class."""

    class TestCommand(object):
        """Tests for the command getter and setter."""

        def test_getter_serial(self):
            """adds -auto to serial tests."""
            test = PiglitGLTest(['foo'])
            assert '-auto' in test.command

        def test_getter_concurrent(self):
            """adds -fbo and -auto to concurrent tests."""
            test = PiglitGLTest(['foo'], run_concurrent=True)
            assert '-auto' in test.command
            assert '-fbo' in test.command

        def test_setter_no_add_auto(self):
            """Doesn't add -fbo or -auto when setting."""
            test = PiglitGLTest(['foo'], run_concurrent=True)
            test.command += ['bar']
            assert '-auto' not in test._command
            assert '-fbo' not in test._command

    class TestIsSkip(object):
        """Tests for the is_skip method and the constructor logic to make it
        work.
        """

        @pytest.yield_fixture()
        def mock_options(self):
            with mock.patch('framework.test.piglit_test.options.OPTIONS',
                            new_callable=Options) as m:
                yield m

        def test_include_and_exclude(self):
            """ raises if include and exclude are given."""
            with pytest.raises(AssertionError):
                PiglitGLTest(['foo'],
                             require_platforms=['glx'],
                             exclude_platforms=['gbm'])

        def test_platform_in_require(self, mock_options):
            """does not skip if platform is in require_platforms."""
            mock_options.env['PIGLIT_PLATFORM'] = 'glx'
            test = PiglitGLTest(['foo'], require_platforms=['glx'])
            test.is_skip()

        def test_platform_not_in_require(self, mock_options):
            """skips if platform is not in require_platforms."""
            mock_options.env['PIGLIT_PLATFORM'] = 'gbm'
            test = PiglitGLTest(['foo'], require_platforms=['glx'])
            with pytest.raises(_TestIsSkip):
                test.is_skip()

        def test_platform_in_exclude(self, mock_options):
            """skips if platform is in exclude_platforms."""
            mock_options.env['PIGLIT_PLATFORM'] = 'glx'
            test = PiglitGLTest(['foo'], exclude_platforms=['glx'])
            with pytest.raises(_TestIsSkip):
                test.is_skip()

        def test_platform_not_in_exclude(self, mock_options):
            """does not skip if platform is in exclude_platforms."""
            mock_options.env['PIGLIT_PLATFORM'] = 'gbm'
            test = PiglitGLTest(['foo'], exclude_platforms=['glx'])
            test.is_skip()



# coding=utf-8
# Copyright (c) 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the monitoring module.

This provides tests for the framework.monitoring modules.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pytest
import six

from framework import exceptions
from framework import monitoring
from . import skip

# pylint: disable=no-self-use,attribute-defined-outside-init


class TestMonitoring(object):
    """Tests for Monitoring methods."""

    def setup(self):
        """Setup for TestMonitoring.

        This create a monitoring.Monitoring instance with monitoring disabled
        to avoid reading the rules in piglit.conf.
        """
        self.regex = r'\*ERROR\*|BUG:'
        self.init_contents = r'foo bar\n'
        self.no_error_contents = r'foo bar\n'
        self.error_contents = r'BUG:bar\n'
        self.monitoring = monitoring.Monitoring(False)

    @skip.linux
    def test_delete_rule(self, tmpdir):
        """monitoring.Monitoring: add and delete rule."""
        p = tmpdir.join('foo')
        p.write(self.init_contents)
        self.monitoring.add_rule('error_file',
                                 'file',
                                 six.text_type(p),
                                 self.regex)
        self.monitoring.update_monitoring()

        self.monitoring.delete_rule('error_file')
        p.write(self.error_contents)
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is False

    def test_add_rule_bad_format(self, tmpdir):
        """monitoring.Monitoring: add non existing type rule."""
        p = tmpdir.join('foo')

        with pytest.raises(exceptions.PiglitFatalError):
            self.monitoring.add_rule('error_file_bad_type',
                                     'bad_type',
                                     six.text_type(p),
                                     self.regex)

    @skip.linux
    def test_file_error(self, tmpdir):
        """monitoring.Monitoring: error found on a file."""
        p = tmpdir.join('foo')
        p.write(self.init_contents)
        self.monitoring.add_rule('error_file',
                                 'file',
                                 six.text_type(p),
                                 self.regex)
        self.monitoring.update_monitoring()

        p.write(self.error_contents)
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is True

    @skip.linux
    def test_file_no_error(self, tmpdir):
        """monitoring.Monitoring: no error found on a file."""
        p = tmpdir.join('foo')
        p.write(self.init_contents)
        self.monitoring.add_rule('no_error_file',
                                 'file',
                                 six.text_type(p),
                                 self.regex)
        self.monitoring.update_monitoring()

        p.write(self.no_error_contents)
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is False

    @skip.linux
    def test_locked_file_error(self, tmpdir):
        """monitoring.Monitoring: error found on a locked file."""
        p = tmpdir.join('foo')
        p.write(self.init_contents)
        self.monitoring.add_rule('error_locked_file',
                                 'locked_file',
                                 six.text_type(p),
                                 self.regex)
        self.monitoring.update_monitoring()

        p.write(self.error_contents)
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is True

    @skip.linux
    def test_locked_file_no_error(self, tmpdir):
        """monitoring.Monitoring: no error found on a locked file."""
        p = tmpdir.join('foo')
        p.write(self.init_contents)
        self.monitoring.add_rule('no_error_file',
                                 'locked_file',
                                 six.text_type(p),
                                 self.regex)
        self.monitoring.update_monitoring()

        p.write(self.no_error_contents)
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is False

    @skip.linux
    def test_dmesg_error(self, mocker):
        """monitoring.Monitoring: error found on the dmesg."""
        mocker.patch('framework.dmesg.subprocess.check_output',
                     mocker.Mock(return_value=b'[1.0]This\n[2.0]is\n[3.0]dmesg'))
        self.monitoring.add_rule('no_error_file',
                                 'dmesg',
                                 '--level emerg,alert,crit,err',
                                 self.regex)
        self.monitoring.update_monitoring()

        mocker.patch('framework.dmesg.subprocess.check_output',
                     mocker.Mock(return_value=b'[4.0]foo\n[5.0]*ERROR* bar'))
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is True

    @skip.linux
    def test_dmesg_no_error(self, mocker):
        """monitoring.Monitoring: no error found on the dmesg."""
        mocker.patch('framework.dmesg.subprocess.check_output',
                     mocker.Mock(return_value=b'[1.0]This\n[2.0]is\n[3.0]dmesg'))
        self.monitoring.add_rule('no_error_file',
                                 'dmesg',
                                 '--level emerg,alert,crit,err',
                                 self.regex)
        self.monitoring.update_monitoring()

        mocker.patch('framework.dmesg.subprocess.check_output',
                     mocker.Mock(return_value=b'[4.0]foo\n[5.0] bar'))
        self.monitoring.check_monitoring()

        assert self.monitoring.abort_needed is False



# encoding=utf-8
# Copyright © 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Classes describing various GLSL types.

These classes do not contain values themselves, rather they describe the GLSL
types.

Included are representations of Matrices, Vectors, and Scalars as containers,
and Ints, Uints, Floats, and Doubles as types. This can easily be expanded to
include more types, and the corresponding containers.

Although the classes that implement these representations are not marked as
private, most consumers will be happier using the provided constants than using
the classes themselves. Probably the only real use for the types is to do type
checking.

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import six

from .utils import lazy_property


class Container(object):
    """A Base class for GLSL container.

    It provides the following attributes:
    name    -- The string formatted name of the container
    scalar  -- True if the type is scalar
    vector  -- True if the type is vector
    matrix  -- True if the type is matrix
    type    -- A Type object that is the type contained by the container
    rows    -- The number of rows the container has. Scalars return 1
    columns -- The number of rows container has. Scalars and Vectors return 1

    """
    def __init__(self, name, is_scalar=False, is_vector=False, is_matrix=False,
                 rows=1, columns=1, contains=None):
        assert all(isinstance(x, bool)
                   for x in [is_scalar, is_vector, is_matrix]), \
            'is_scalar, is_vector, and is_matrix must be bools'
        assert [is_scalar, is_vector, is_matrix].count(True) == 1, \
            'A type must be exactly one of: scalar, vector, matrix'
        assert isinstance(contains, Type), 'contains must be a type instance'
        assert isinstance(name, six.text_type), 'name must be string'
        assert columns is None or isinstance(columns, int), \
            'columns must be an int if provided'
        assert rows is None or isinstance(rows, int), \
            'row must be an int if provided'

        self.name = name
        self.scalar = is_scalar
        self.vector = is_vector
        self.matrix = is_matrix
        self.type = contains
        self.rows = rows
        self.columns = columns


class Scalar(Container):
    """A base class for Scalar types."""
    def __init__(self, name, contains):
        super(Scalar, self).__init__(name, is_scalar=True, contains=contains)


class Vector(Container):
    """A base class for Vector Containers."""
    def __init__(self, name, rows, contains):
        assert isinstance(rows, int), 'rows must be an integer'
        assert rows in [2, 3, 4], 'vecs can only be 2-4 in GLSL'

        super(Vector, self).__init__(name, is_vector=True, rows=rows,
                                     contains=contains)


class Matrix(Container):
    """A base class for vectory Containers."""
    def __init__(self, name, columns, rows, contains):
        assert isinstance(rows, int), 'rows must be an integer'
        assert isinstance(columns, int), 'columns must be an integer'
        assert rows in [2, 3, 4], 'Matrix rows can only be 2-4 in GLSL'
        assert columns in [2, 3, 4], 'Matrix columns can only be 2-4 in GLSL'

        super(Matrix, self).__init__(name, is_matrix=True, rows=rows,
                                     columns=columns, contains=contains)

    @lazy_property
    def square(self):
        return self.columns == self.rows


class Type(object):
    """Class representing a GLSL type.

    Provides the following interfaces currently:
    integer -- True if the type is an integer type
    signed  -- True if the type is a signed type
    float   -- True if the type is a floating point type
    size    -- The integer size of the type (8, 16, 32, 64, 128, etc)
    name    -- A formated string name of the type. (float, double, int64, ...)

    """
    def __init__(self, name, integer=False, signed=False, floating=False,
                 size=None):
        assert [integer, floating].count(True) == 1, \
            'Type can onnly be float or int'
        assert integer or (float and signed), 'Floats cannot be unsigned'
        assert isinstance(size, int), 'size must be an int'

        self.name = name
        self.integer = integer
        self.signed = signed
        self.float = floating
        self.size = size


# pylint: disable=bad-whitespace
# Type definitions, these are used internally by the Scalar/Vector/Matrix
# constructs
INT_TYPE    = Type('int',    integer=True,  signed=True,  size=32)
UINT_TYPE   = Type('uint',   integer=True,  signed=False, size=32)
FLOAT_TYPE  = Type('float',  floating=True, signed=True,  size=32)
DOUBLE_TYPE = Type('double', floating=True, signed=True,  size=64)

# Container definitions
INT     = Scalar('int',    INT_TYPE)
UINT    = Scalar('uint',   UINT_TYPE)
FLOAT   = Scalar('float',  FLOAT_TYPE)
DOUBLE  = Scalar('double', DOUBLE_TYPE)

VEC2    = Vector('vec2',  2, FLOAT_TYPE)
VEC3    = Vector('vec3',  3, FLOAT_TYPE)
VEC4    = Vector('vec4',  4, FLOAT_TYPE)

IVEC2   = Vector('ivec2', 2, INT_TYPE)
IVEC3   = Vector('ivec3', 3, INT_TYPE)
IVEC4   = Vector('ivec4', 4, INT_TYPE)

UVEC2   = Vector('uvec2', 2, UINT_TYPE)
UVEC3   = Vector('uvec3', 3, UINT_TYPE)
UVEC4   = Vector('uvec4', 4, UINT_TYPE)

DVEC2   = Vector('dvec2', 2, DOUBLE_TYPE)
DVEC3   = Vector('dvec3', 3, DOUBLE_TYPE)
DVEC4   = Vector('dvec4', 4, DOUBLE_TYPE)

MAT2    = Matrix('mat2',    2, 2, FLOAT_TYPE)
MAT3    = Matrix('mat3',    3, 3, FLOAT_TYPE)
MAT4    = Matrix('mat4',    4, 4, FLOAT_TYPE)
MAT2X2  = Matrix('mat2x2',  2, 2, FLOAT_TYPE)
MAT2X3  = Matrix('mat2x3',  2, 3, FLOAT_TYPE)
MAT2X4  = Matrix('mat2x4',  2, 4, FLOAT_TYPE)
MAT3X2  = Matrix('mat3x2',  3, 2, FLOAT_TYPE)
MAT3X3  = Matrix('mat3x3',  3, 3, FLOAT_TYPE)
MAT3X4  = Matrix('mat3x4',  3, 4, FLOAT_TYPE)
MAT4X2  = Matrix('mat4x2',  4, 2, FLOAT_TYPE)
MAT4X3  = Matrix('mat4x3',  4, 3, FLOAT_TYPE)
MAT4X4  = Matrix('mat4x4',  4, 4, FLOAT_TYPE)

DMAT2   = Matrix('dmat2',   2, 2, DOUBLE_TYPE)
DMAT3   = Matrix('dmat3',   3, 3, DOUBLE_TYPE)
DMAT4   = Matrix('dmat4',   4, 4, DOUBLE_TYPE)
DMAT2X2 = Matrix('dmat2x2', 2, 2, DOUBLE_TYPE)
DMAT2X3 = Matrix('dmat2x3', 2, 3, DOUBLE_TYPE)
DMAT2X4 = Matrix('dmat2x4', 2, 4, DOUBLE_TYPE)
DMAT3X2 = Matrix('dmat3x2', 3, 2, DOUBLE_TYPE)
DMAT3X3 = Matrix('dmat3x3', 3, 3, DOUBLE_TYPE)
DMAT3X4 = Matrix('dmat3x4', 3, 4, DOUBLE_TYPE)
DMAT4X2 = Matrix('dmat4x2', 4, 2, DOUBLE_TYPE)
DMAT4X3 = Matrix('dmat4x3', 4, 3, DOUBLE_TYPE)
DMAT4X4 = Matrix('dmat4x4', 4, 4, DOUBLE_TYPE)
# pylint: enable=bad-whitespace



# coding=utf-8
# Copyright 2013 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Authors: Tom Stellard <thomas.stellard@amd.com>
#          Aaron Watry  <awatry@gmail.com>
#

from __future__ import print_function, division, absolute_import
import os

from genclbuiltins import gen, NEGNAN
from math import radians, degrees, pi

CLC_VERSION_MIN = {
    'clamp' : 10,
    'degrees' : 10,
    'max' : 10,
    'min' : 10,
    'mix' : 10,
    'radians' : 10,
    'step' : 10,
    'smoothstep' : 10,
    'sign' : 10
}

DATA_TYPES = ['float']

F = {
    'float' : 'float'
}

I = {
    'float' : 'int'
}

tests = {
    'clamp' : {
        'arg_types': [F, F, F, F],
        'function_type': 'tss',
        'values': [
            [0.5,  0.0, 0.0,  0.0, float("nan")], #Result
            [1.0, -0.5, 0.0,  0.0, float("nan")], #Arg0
            [0.0,  0.0, 0.0, -0.5, float("nan")], #Arg1
            [0.5,  0.5, 0.0,  0.5, float("nan")], #Arg2
        ]
    },
    'degrees' : {
        'arg_types': [F, F],
        'function_type': 'ttt',
        'values': [
            [degrees(0.5), degrees(-0.5), 180.0, 0.0, 360, 1800.0, 18000, 90], #Result
            [0.5,                   -0.5, pi,    0.0, 2*pi, 10*pi, 100*pi, pi/2] #Arg0
        ]
    },
    'max' : {
        'arg_types': [F, F, F],
        'function_type': 'tss',
        'values': [
            [1.0,  0.0, 0.0,  0.0 ], #Result
            [1.0, -0.5, 0.0,  0.0 ], #Arg0
            [0.0,  0.0, 0.0, -0.5 ] #Arg1
        ]
    },
    'min' : {
        'arg_types': [F, F, F],
        'function_type': 'tss',
        'values': [
            [0.0, -0.5, 0.0, -0.5 ], #Result
            [1.0, -0.5, 0.0,  0.0 ], #Arg0
            [0.0,  0.0, 0.0, -0.5 ] #Arg1
        ]
    },
    'mix' : { #x + (y - x) * a
        'arg_types': [F, F, F, F],
        'function_type': 'tts',
        'values': [
            [0.5, -0.25, 0.0, -0.5, 1.0, 3.0, 10.0, float("nan"), float("nan"), float("nan")], #Result
            [1.0, -0.5,  0.0,  0.0, 1.0, 4.0, 15.0, 5.0, 4.0, float("nan")], #Arg0
            [0.0,  0.0,  0.0, -0.5, 2.0, 2.0, 10.0, -0.2, float("nan"), 1.5], #Arg1
            [0.5,  0.5,  0.0,  1.0, 0.0, 0.5,  1.0, float("nan"), 1.0, 0.0], #Arg2
        ]
    },
    'radians' : {
        'arg_types': [F, F],
        'function_type': 'ttt',
        'values': [
            [0.5,                   -0.5, pi,    0.0, 2*pi, 10*pi, 100*pi, pi/2], #Result
            [degrees(0.5), degrees(-0.5), 180.0, 0.0, 360, 1800.0, 18000, 90] #Arg0
        ]
    },
    #TODO Add scalar combination (tst?)
    'step' : {
        'arg_types': [F, F, F],
        'function_type': 'ttt',
        'values': [
            [0.0,  1.0, 1.0,  0.0, 1.0, 1.0], #Result
            [1.0, -0.5, 0.0,  0.0, 1.0, float("nan")], #Arg0
            [0.0,  0.0, 0.0, -0.5, float("nan"), 1.0] #Arg1
        ]
    },
    #TODO Add scalar combination (tst?)
    'smoothstep' : {
        'arg_types': [F, F, F, F],
        'function_type': 'ttt',
        'values': [
            [0.0,  0.0, 1.0,  1.0, 0.5,  0.896], #Result
            [0.0,  0.0, 0.0, -0.5, 0.0,  0.0], #Arg0
            [1.0,  1.0, 0.5,  0.0, 0.5,  0.5], #Arg1
            [-0.5, 0.0, 0.5,  1.0, 0.25, 0.4] #Arg2
        ]
    },
    'sign' : {
        'arg_types': [F, F],
        'function_type': 'ttt',
        'values': [
            [1.0, -1.0, 0.0, -0.0, 0.0], #Result
            [0.5, -0.5, 0.0, -0.0, float("nan")] #Arg0
        ]
    }
}


def main():
    dirName = os.path.join("cl", "builtin", "common")

    testDefs = {}
    functions = sorted(tests.keys())
    for dataType in DATA_TYPES:
        for fnName in functions:
            testDefs[(dataType, fnName)] = tests[fnName]

    gen(DATA_TYPES, CLC_VERSION_MIN, functions, testDefs, dirName)


if __name__ == '__main__':
    main()
